package eu.jonahbauer.raytracing;

import eu.jonahbauer.raytracing.material.DielectricMaterial;
import eu.jonahbauer.raytracing.material.LambertianMaterial;
import eu.jonahbauer.raytracing.material.Material;
import eu.jonahbauer.raytracing.material.MetallicMaterial;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.render.*;
import eu.jonahbauer.raytracing.render.Color;
import eu.jonahbauer.raytracing.render.canvas.BufferedImageCanvas;
import eu.jonahbauer.raytracing.scene.Hittable;
import eu.jonahbauer.raytracing.scene.Scene;
import eu.jonahbauer.raytracing.scene.Sphere;
import org.jetbrains.annotations.NotNull;

import javax.swing.*;
import java.awt.*;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.nio.file.Path;
import java.util.ArrayList;

public class Main {
    public static void main(String[] args) throws IOException {
        var scene = getScene();
        var camera = Camera.builder()
//                .withImage(1920, 1080)
                .withImage(800, 450)
                .withPosition(new Vec3(13, 2, 3))
                .withTarget(new Vec3(0, 0, 0))
                .withSamplesPerPixel(100)
                .withMaxDepth(10)
                .withFieldOfView(Math.toRadians(20))
                .withBlurAngle(Math.toRadians(0.6))
                .withFocusDistance(10.0)
                .build();

        var canvas = new BufferedImageCanvas(camera.width(), camera.height());
        preview(canvas);

        camera.render(scene, canvas, Camera.RenderMode.SAMPLES);
        ImageFormat.PNG.write(canvas, Path.of("scene-" + System.currentTimeMillis() + ".png"));
    }

    private static Thread preview(@NotNull BufferedImageCanvas canvas) {
        var frame = new JFrame();
        frame.setSize(canvas.width(), canvas.height());
        frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        frame.setContentPane(new JPanel() {
            @Override
            protected void paintComponent(Graphics g) {
                g.drawImage(canvas.getImage(), 0, 0, null);
            }
        });
        frame.setResizable(false);
        frame.setVisible(true);

        var update = Thread.ofVirtual().start(() -> {
            while (!Thread.interrupted()) {
                try {
                    Thread.sleep(1000);
                } catch (InterruptedException ex) {
                    break;
                }
                frame.repaint();
            }
        });

        frame.addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent e) {
                update.interrupt();
            }
        });

        return update;
    }

    private static @NotNull Scene getScene() {
        var objects = new ArrayList<Hittable>();
        objects.add(new Sphere(new Vec3(0, -1000, 0), 1000, new LambertianMaterial(new Color(0.5, 0.5, 0.5))));

        for (int a = -11; a < 11; a++) {
            for (int b = -11; b < 11; b++) {
                var center = new Vec3(a + 0.9 * Math.random(), 0.2, b + 0.9 * Math.random());
                if (Vec3.distance(center, new Vec3(4, 0.2, 0)) <= 0.9) continue;

                Material material;
                var rnd = Math.random();
                if (rnd < 0.8) {
                    // diffuse
                    var albedo = Color.multiply(Color.random(), Color.random());
                    material = new LambertianMaterial(albedo);
                } else if (rnd < 0.95) {
                    // metal
                    var albedo = Color.random(0.5, 1.0);
                    var fuzz = Math.random() * 0.5;
                    material = new MetallicMaterial(albedo, fuzz);
                } else {
                    // glass
                    material = new DielectricMaterial(1.5);
                }

                objects.add(new Sphere(center, 0.2, material));
            }
        }

        objects.add(new Sphere(new Vec3(0, 1, 0), 1.0, new DielectricMaterial(1.5)));
        objects.add(new Sphere(new Vec3(-4, 1, 0), 1.0, new LambertianMaterial(new Color(0.4, 0.2, 0.1))));
        objects.add(new Sphere(new Vec3(4, 1, 0), 1.0, new MetallicMaterial(new Color(0.7, 0.6, 0.5))));

        return new Scene(objects);
    }
}