package eu.jonahbauer.raytracing.render;

import eu.jonahbauer.raytracing.math.Range;
import eu.jonahbauer.raytracing.math.Ray;
import eu.jonahbauer.raytracing.math.Vec3;
import eu.jonahbauer.raytracing.render.canvas.Canvas;
import eu.jonahbauer.raytracing.render.canvas.Image;
import eu.jonahbauer.raytracing.scene.Scene;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Objects;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Function;
import java.util.stream.IntStream;
import java.util.stream.LongStream;

public final class Camera {
    // image size
    private final int width;
    private final int height;

    // camera position and orientation
    private final @NotNull Vec3 origin;

    // rendering
    private final int samplesPerPixel;
    private final int maxDepth;
    private final double gamma;
    private final double blurRadius;

    // internal properties
    private final @NotNull Vec3 u;
    private final @NotNull Vec3 v;

    private final @NotNull Vec3 pixelU;
    private final @NotNull Vec3 pixelV;
    private final @NotNull Vec3 pixel00;

    public static @NotNull Builder builder() {
        return new Builder();
    }

    public Camera() {
        this(new Builder());
    }

    private Camera(@NotNull Builder builder) {
        this.width = builder.imageWidth;
        this.height = builder.imageHeight;

        var viewportHeight = 2 * Math.tan(0.5 * builder.fov) * builder.focusDistance;
        var viewportWidth = viewportHeight * ((double) width / height);

        this.origin = builder.position;
        var direction = (builder.direction == null ? builder.target.minus(builder.position).unit() : builder.direction);

        this.samplesPerPixel = builder.samplePerPixel;
        this.maxDepth = builder.maxDepth;
        this.gamma = builder.gamma;
        this.blurRadius = Math.tan(0.5 * builder.blurAngle) * builder.focusDistance;

        // project direction the horizontal plane
        var dXZ = direction.withY(0).unit();
        this.u = Vec3.rotate(
                new Vec3(- dXZ.z(), 0, dXZ.x()), // perpendicular to d in horizontal plane
                direction, builder.rotation
        );
        this.v = direction.cross(u);  // perpendicular to viewportU and direction

        this.pixelU = u.times(viewportWidth / width);
        this.pixelV = v.times(viewportHeight / height);

        this.pixel00 = origin.plus(direction.times(builder.focusDistance))
                .minus(u.times(0.5 * viewportWidth)).minus(v.times(0.5 * viewportHeight))
                .plus(pixelU.div(2)).plus(pixelV.div(2));
    }

    public @NotNull Image render(@NotNull Scene scene) {
        var image = new Image(width, height);
        render(scene, image);
        return image;
    }

    public void render(@NotNull Scene scene, @NotNull Canvas canvas) {
        render(scene, canvas, RenderMode.SEQUENTIEL);
    }

    public void render(@NotNull Scene scene, @NotNull Canvas canvas, @NotNull RenderMode mode) {
        if (canvas.width() != width || canvas.height() != height) throw new IllegalArgumentException();

        var progress = new AtomicInteger();
        mode.render(this, scene, canvas, () -> {
            var val = progress.incrementAndGet();
            if (val % 1000 == 0) System.out.println(val);
        });
    }

    private @NotNull Ray getRay(int x, int y) {
        var origin = this.origin;
        if (blurRadius > 0) {
            double bu, bv;
            do {
                bu = 2 * Math.random() - 1;
                bv = 2 * Math.random() - 1;
            } while (bu * bu + bv * bv >= 1);

            origin = origin.plus(u.times(blurRadius * bu)).plus(v.times(blurRadius * bv));
        }

        return new Ray(origin, getPixel(x, y).minus(origin));
    }

    private @NotNull Vec3 getPixel(int x, int y) {
        Objects.checkIndex(x, width);
        Objects.checkIndex(y, height);

        double dx = x + Math.random() - 0.5;
        double dy = y + Math.random() - 0.5;
        return pixel00.plus(pixelU.times(dx)).plus(pixelV.times(dy));
    }

    private @NotNull Color getColor(@NotNull Scene scene, @NotNull Ray ray) {
        return getColor(scene, ray, maxDepth);
    }

    private @NotNull Color getColor(@NotNull Scene scene, @NotNull Ray ray, int depth) {
        if (depth <= 0) return Color.BLACK;

        var optional = scene.hit(ray, new Range(0.001, Double.POSITIVE_INFINITY));
        if (optional.isPresent()) {
            var hit = optional.get();
            var material = hit.material();
            return material.scatter(ray, hit)
                    .map(scatter -> Color.multiply(scatter.attenuation(), getColor(scene, scatter.ray(), depth - 1)))
                    .orElse(Color.BLACK);
        } else {
            return getSkyboxColor(ray);
        }
    }

    private static @NotNull Color getSkyboxColor(@NotNull Ray ray) {
        // altitude from -pi/2 to pi/2
        var alt = Math.copySign(
                Math.acos(ray.direction().withY(0).unit().times(ray.direction().unit())),
                ray.direction().y()
        );
        return Color.lerp(Color.WHITE, Color.SKY, alt / Math.PI + 0.5);
    }

    public int width() {
        return width;
    }

    public int height() {
        return height;
    }

    public enum RenderMode {
        SEQUENTIEL {
            @Override
            public void render(@NotNull Camera camera, @NotNull Scene scene, @NotNull Canvas canvas, @NotNull Runnable onProgressUpdate) {
                for (int y = 0; y < camera.height; y++) {
                    for (int x = 0; x < camera.width; x++) {
                        Color color = Color.BLACK;
                        for (int i = 1; i <= camera.samplesPerPixel; i++) {
                            var ray = camera.getRay(x, y);
                            var c = camera.getColor(scene, ray);
                            color = Color.average(color, c, i);
                        }
                        canvas.set(x, y, color);
                        onProgressUpdate.run();
                    }
                }
            }
        },
        PARALLEL {
            @Override
            public void render(@NotNull Camera camera, @NotNull Scene scene, @NotNull Canvas canvas, @NotNull Runnable onProgressUpdate) {
                coordinates(camera.width, camera.height).parallel().forEach(pos -> {
                    var x = (int) pos;
                    var y = (int) (pos >> 32);
                    Color color = Color.BLACK;
                    for (int i = 1; i <= camera.samplesPerPixel; i++) {
                        var ray = camera.getRay(x, y);
                        var c = camera.getColor(scene, ray);
                        color = Color.average(color, c, i);
                    }
                    canvas.set(x, y, color);
                    onProgressUpdate.run();
                });
            }
        },
        SAMPLES {
            @Override
            public void render(@NotNull Camera camera, @NotNull Scene scene, @NotNull Canvas canvas, @NotNull Runnable onProgressUpdate) {
                for (int i = 1; i <= camera.samplesPerPixel; i++) {
                    var sample = i;
                    coordinates(camera.width, camera.height).forEach(pos -> {
                        var x = (int) pos;
                        var y = (int) (pos >> 32);
                        var ray = camera.getRay(x, y);
                        var color = Color.average(canvas.get(x, y), camera.getColor(scene, ray), sample);
                        canvas.set(x, y, color);
                    });
                    onProgressUpdate.run();
                }
            }
        }
        ;

        private static LongStream coordinates(int width, int height) {
            return IntStream.range(0, height)
                    .mapToObj(y -> IntStream.range(0, width).mapToLong(x -> (long) y << 32 | x))
                    .flatMapToLong(Function.identity());
        }

        public abstract void render(@NotNull Camera camera, @NotNull Scene scene, @NotNull Canvas canvas, @NotNull Runnable onProgressUpdate);

    }

    public static class Builder {
        private int imageWidth = 1920;
        private int imageHeight = 1080;

        private @NotNull Vec3 position = Vec3.ZERO;
        private @Nullable Vec3 direction = Vec3.UNIT_Z.neg();
        private @Nullable Vec3 target = null;
        private double rotation = 0.0;

        private double fov = 0.5 * Math.PI;
        private double focusDistance = 10;
        private double blurAngle = 0.0;

        private int samplePerPixel = 100;
        private int maxDepth = 10;
        private double gamma = 2.0;

        private Builder() {}

        public @NotNull Builder withImage(int width, int height) {
            if (width <= 0 || height <= 0) throw new IllegalArgumentException("width and height must be positive");
            this.imageWidth = width;
            this.imageHeight = height;
            return this;
        }

        public @NotNull Builder withPosition(@NotNull Vec3 position) {
            this.position = Objects.requireNonNull(position);
            return this;
        }

        public @NotNull Builder withDirection(@NotNull Vec3 direction) {
            this.direction = Objects.requireNonNull(direction).unit();
            this.target = null;
            return this;
        }

        public @NotNull Builder withTarget(@NotNull Vec3 target) {
            this.target = Objects.requireNonNull(target);
            this.direction = null;
            return this;
        }

        public @NotNull Builder withRotation(double rotation) {
            if (!Double.isFinite(rotation)) throw new IllegalArgumentException("rotation must be finite");
            this.rotation = rotation;
            return this;
        }

        public @NotNull Builder withFieldOfView(double fov) {
            if (fov <= 0 || fov >= Math.PI || !Double.isFinite(fov)) throw new IllegalArgumentException("fov must be in the range (0, π)");
            this.fov = fov;
            return this;
        }

        public @NotNull Builder withFocusDistance(double focusDistance) {
            if (focusDistance <= 0 || !Double.isFinite(focusDistance)) throw new IllegalArgumentException("focus distance must be positive");
            this.focusDistance = focusDistance;
            return this;
        }

        public @NotNull Builder withBlurAngle(double angle) {
            if (angle < 0 || angle >= Math.PI || !Double.isFinite(angle)) throw new IllegalArgumentException("blur-angle must be in the range [0, π)");
            this.blurAngle = angle;
            return this;
        }

        public @NotNull Builder withSamplesPerPixel(int samples) {
            if (samples <= 0) throw new IllegalArgumentException("samples must be positive");
            this.samplePerPixel = samples;
            return this;
        }

        public @NotNull Builder withMaxDepth(int depth) {
            if (depth <= 0) throw new IllegalArgumentException("depth must be positive");
            this.maxDepth = depth;
            return this;
        }

        public @NotNull Builder withGamma(double gamma) {
            if (gamma <= 0 || !Double.isFinite(gamma)) throw new IllegalArgumentException("gamma must be positive");
            this.gamma = gamma;
            return this;
        }

        public @NotNull Camera build() {
            return new Camera(this);
        }

    }
}
